from numpy import linspace
import matplotlib.pyplot as plt
from sklearn.metrics import roc_curve, roc_auc_score
from .. import plot as ndplot
from . import predict as ndmlpred


def plot_pred_comparison_by_track_property(dataset, pred_name,
        pred_comparison, bin_property, bins=10, threshold=.6,
        legend_id=None, ax=None):
    """Compares true labels to the model predictions by some function,
    binned by a track property present in data.

    Args:
        dataset: a TrackPropertiesDataset.
        pred_name: the name of a prediction to be found in dataset
        pred_comparison: a function that takes in the labels, the
            predicted labels, and a threshold value, and returns a
            number measuring some property of the predicted labels'
            relation to the actual ones.
        bin_property: a property in data_properties or the
            label_property that will split the dataset into bins.
        bins: either an int for the number of bins, a 3-tuple of the
            form (low_bound, high_bound, num_bins), or a list of
            numbers. See ntupledict.operations.make_bins() for info.
        threshold: the limit at which a prediction signifies one or
            the other value of a binary classification,
        legend_id: the entry in the legend for the line to be plotted.
            Calling ax.legend() should be done outside this function.
        ax: an axes object to be used to plot in this function.

    Returns:
        The Axes object to be used to plot in this function.
    """

    if ax is None:
        ax = plt.figure().add_subplot(111)

    track_prop_dict = dataset.to_track_prop_dict(include_preds=True)

    def measure_pred_comparison(track_prop_dict):
        """Measures some prediction comparison of true labels and
        predicted labels that are within a track propeties dict given
        the input function to the parent function.

        Returns:
            A tuple containing the measured prediction comparison and
            the error."""

        return pred_comparison(track_prop_dict[dataset.get_label_property()],
                track_prop_dict[pred_name], threshold)

    return ndplot.plot_measure_by_bin(track_prop_dict, bin_property,
            measure_pred_comparison, bins, legend_id, ax)


def plot_pred_comparison_by_threshold(dataset, pred_name,
        pred_comparison, thresholds=10, legend_id=None,
        ax=None):
    """Compares true labels to the model predictions by some function
    at various thresholds.

    Args:
        dataset: a TrackPropertiesDataset.
        pred_name: the name of a prediction to be found in dataset.
        pred_comparison: a function that takes in the labels, the
            predicted labels, and a threshold value, and returns a
            number measuring some property of the predicted labels'
            relation to the actual ones.
        thresholds: the limits at which a prediction signifies one or
            the other value of a binary classification.
        legend_id: the entry in the legend for the line to be plotted.
            Calling ax.legend() should be done outside this function.
        ax: an axes object to be used to plot in this function.

    Returns:
        The axes object to used to plot in this function.
    """

    if ax is None:
        ax = plt.figure().add_subplot(111)

    # Generate threshold list if thresholds is not a list
    if not isinstance(thresholds, list):
        thresholds = linspace(0, 1, thresholds)

    labels = dataset.get_labels()
    predictions = dataset.get_prediction(pred_name)

    pred_comps_with_errors = list(map(lambda threshold:
        pred_comparison(labels, predictions, threshold),
        thresholds))

    pred_comps = list(map(lambda l: l[0], pred_comps_with_errors))
    pred_errs = list(map(lambda l: l[1], pred_comps_with_errors))

    ax.errorbar(thresholds, pred_comps, pred_errs,
        label=legend_id, fmt=".")
    ax.set_xlabel("Decision threshold")

    return ax


def plot_rocs(dataset, prob_pred_names=[], def_pred_names=[],
              xlims=(0, .3), ylims=(.9, 1)):
    """Create ROC curves through true positive rate / false positive
    rate space for different models by changing the cut on predictions.
    Note that this only works if the label is from a binary classifier
    such as trk_genuine.

    Args:
        dataset: a TrackPropertiesDataset containing the data, labels,
            and corresponding property names for both.
        prob_pred_names: names of probablistic predictions accessible
            from dataset. Typically the names of the models that made
            them. Will be plotted as a curve.
        def_pred_names: names of pre-thresholded predictions
            accessible from the dataset. This is what is used for cut-
            generated predictions. Plotted as a point.

    Returns:
        The axes object to used to plot in this function.
    """

    ax = plt.figure().add_subplot(111)

    # Plot ROC curve for models
    labels = dataset.get_labels()
    for prob_pred_name in prob_pred_names:
        pred_prob_labels = dataset.get_prediction(prob_pred_name)
        fpr, tpr, _ = roc_curve(labels, pred_prob_labels)
        auc = roc_auc_score(labels, pred_prob_labels)
        label = "{} ({})".format(prob_pred_name, str(round(auc, 3)))
        ax.plot(fpr, tpr, label=label, linewidth=2)

    # Plot cuts, if any are given
    for def_pred_name in def_pred_names:
        pred_labels = dataset.get_prediction(def_pred_name)
        fpr_cut = ndmlpred.false_positive_rate(labels, pred_labels)
        tpr_cut = ndmlpred.true_positive_rate(labels, pred_labels)
        ax.scatter(fpr_cut, tpr_cut, s=80, marker="*",
                label="cuts: {}".format(def_pred_name), color="red")

    ax.tick_params(labelsize=14)
    ax.set_xlabel("FPR", fontsize=20)
    ax.set_ylabel("TPR", fontsize=20)
    ax.set_xlim(xlims)
    ax.set_ylim(ylims)
    ax.legend(loc="best", fontsize=14)

    return ax



"""LOAD: loads root files into ntuple dicts."""


from uproot import open as uproot_open
from . import operations as ndops
from .operations import select as sel


def root_files_to_ntuple_dict(root_ntuple_paths, properties_by_track_type,
        keep_invalid_vals=False):
    """The first function to be run in a typical ntupledicts session.

    Takes in paths to root track-trigger ntuples and a dict from
    track types to desired properties to be included, returns an ntuple
    dictionary formed by selecting properties from the ntuples and then
    concatenating them all together. Cuts any invalid values, like inf
    or nan, by default.

    Args:
        root_ntuple_paths: an iterable of paths to track-trigger root
            ntuples.
        properties_by_track_type: a dictionary from track types (trk,
            matchtrk, etc.) to properties to be selected
            (eta, pt, chi2).
        keep_invalid_vals: if True, don't cut tracks with inf or nan as
            one of their values.

    Returns:
        An ntuple dict, the star of the show.

    Raises:
        IOError: if any of the root files cannot be opened by uproot.
        IOError: if an event tree cannot be read from any opened uproot
            files.
    """

    # Extract uproot event trees
    uproot_ntuples = []
    for root_ntuple_path in root_ntuple_paths:

        # Open root files using uproot
        try:
            uprooted_file = uproot_open(root_ntuple_path)
        except:
            raise IOError("Root file at {} cannot be opened by uproot."
                    .format(root_ntuple_path))

        # Access event tree from opened uproot file
        try:
            uproot_event_set = next(iter(uprooted_file.values()))
            uproot_event_tree = next(iter(uproot_event_set.values()))
            uproot_ntuples.append(uproot_event_tree)
        except:
            raise IOError("Event tree cannot be read from opened file at {}."
                    .format(root_ntuple_path))

    return uproot_ntuples_to_ntuple_dict(uproot_ntuples,
            properties_by_track_type,
            keep_invalid_vals)


def uproot_ntuples_to_ntuple_dict(uproot_ntuples, properties_by_track_type,
        keep_invalid_vals=False):
    """Takes in a collection of uproot ntuples and a dictionary from
    track types to desired properties to be included, returns an ntuple
    dictionary formed by selecting properties from the ntuples and then
    concatenating them all together. Cuts any invalid values, like inf
    or nan, by default.

    Args:
        uproot_ntuples: an iterable of uproot ntuples.
        properties_by_track_type: a dictionary from track types (trk,
            matchtrk, etc.) to properties to be selected
            (eta, pt, chi2).
        keep_invalid_vals: if True, don't cut tracks with inf or nan as
            one of their values.

    Returns:
        An ntuple dict.
    """

    return ndops.add_ntuple_dicts(list(map(lambda uproot_ntuple:
        uproot_ntuple_to_ntuple_dict(uproot_ntuple,
            properties_by_track_type, keep_invalid_vals),
        uproot_ntuples)))


def uproot_ntuple_to_ntuple_dict(uproot_ntuple, properties_by_track_type,
        keep_invalid_vals=False):
    """Turns an uproot ntuple into an ntuple dictionary.

    Args:
        uproot_ntuple: an uproot ntuple.
        properties_by_track_type: a dictionary from track types (trk,
             matchtrk, etc.) to properties to be selected
             (eta, pt, chi2).
        keep_invalid_vals: if True, don't cut tracks with inf or nan as
            one of their values.

    Returns:
        An ntuple dict.
    """

    ntuple_dict = dict(map(lambda track_type, track_properties:
        (track_type, uproot_ntuple_to_track_prop_dict(
            uproot_ntuple, track_type, track_properties)),
        properties_by_track_type.keys(), properties_by_track_type.values()))

    if keep_invalid_vals:
        return ntuple_dict
    else:
        invalid_vals = [float("inf"), float("nan")]
        invalid_track_sel = lambda val: val not in invalid_vals

        # Select for the above selector in every field of an ntuple dict,
        # but only if the list contains those values in the first place
        invalid_sel_dict = dict(map(lambda track_type:
            (track_type, dict(map(lambda track_property:
                (track_property, invalid_track_sel),
                filter(lambda track_property:
                    any([invalid_val in ntuple_dict[track_type][track_property]\
                            for invalid_val in invalid_vals]),
                    ntuple_dict[track_type].keys()
                    )))),
            ntuple_dict.keys()))

        return ndops.cut_ntuple_dict(ntuple_dict, invalid_sel_dict)


def uproot_ntuple_to_track_prop_dict(uproot_ntuple, track_type,
        track_properties):
    """Takes in an uproot ntuple, the data type, and properties to be
    extracted; returns a dictionary from a property name to flattened
    array of values. Note that due to this flattening, all information
    about which tracks are from which event is lost.

    Args:
        uproot_ntuple: an uproot ntuple.
        track_type: a track type string. "trk", "matchtrk", etc.
        track_properties: a list of track property strings. "pt",
            "eta", "pdgid", etc.

    Returns:
        A track properties dict.
    """

    def get_value_list(track_property):
        """Returns the value list corresponding to uproot ntuple, track
        type, and property name."""

        return list(uproot_ntuple["{}_{}".format(track_type, track_property)]
                    .array().flatten())

    return dict(map(lambda track_property:
                    (track_property, get_value_list(track_property)),
                    track_properties))





from tensorflow import constant as tfconst
from tensorflow import transpose as tftrans
from tensorflow.keras.utils import normalize as tfnorm
from tensorflow import float64
from .. import operations as ndops
from copy import deepcopy


class TrackPropertiesDataset:
    """A track property dict disguised as a tensorflow dataset.

    The properties of both the data and the label are set dynamically
    such that this dataset can store more property data than a model
    will train on. This is done by introducing "active properties",
    which are a subset of the total data stored.

        print(list(tpd.keys()))  # A track properties dict with keys
                                 # ["pt", "chi2", "nstub", "genuine"]
        # set genuine as a label property
        tpds = TrackPropertiesDataset(tpd, "genuine")
        tpds.get_available_data_properties()  # ["pt", "chi2", "nstub",
                                              # "genuine"]
        # by default, the label property is not an active data property
        tpds.get_active_data_properties()  # ["pt", "chi2", "nstub"]
        tpds.set_active_data_properties(["pt"])
        tpds.get_active_data_properties()  # ["pt"]
        tpds.set_active_data_properties(["chi2", "matchtp_pdgid"])
            # ValueError: matchtp_pdgid not an available data property

    By default, self.get_data() returns data normalized for each track
    property for maximum model compatibility, though this can be
    disabled with the kwarg normalize=False.

    This dataset can also store predictions, accept selector dicts to
    preform cuts, and be split into multiple datasets of the same form
    but of different size.
    """

    def __init__(self, track_prop_dict, label_property,
                 active_data_properties=[], prediction_dict={}):
        """Initializes this TrackPropertiesDataset with its track
        properties dict, its access settings, and predictions, if any.

        Args:
            track_prop_dict: a track properties dict.
            label_property: a track property in track_prop_dict.
            active_data_properties: a list that is a subset of the
                properties in track_prop_dict, or None. If None, all
                available properties become active save for the label
                property.
            prediction_dict: a dictionary from prediction names
                (typically the name of the model) to prediction lists.
        """

        # Determine whether the given track properties dict is valid by using
        # the most basic function checking track properties dict validity
        ndops.track_prop_dict_length(track_prop_dict)

        self._track_prop_dict = track_prop_dict
        self.set_label_property(label_property)
        self.set_active_data_properties(active_data_properties)
        self._predictions = prediction_dict

    def __add__(self, other):
        """Add this TrackPropertiesDataset together with another.

        Args:
            other: a TrackPropertiesDataset that has the same available
                data properties, label property, active data properties,
                and prediction names.

        Returns:
            A TrackPropertiesDataset that has the tracks of this
            dataset and the other one, with the same settings.

        Raises:
            ValueError: if the other dataset has different data
                properties, label property, active data properties, or
                prediction names than this one.
        """

        if self.get_label_property() != other.get_label_property():
            raise ValueError("Chosen label properties do not match.")
        if self.get_active_data_properties() \
                != other.get_active_data_properties():
            raise ValueError("Active data properties do not match.")
        if self.get_available_data_properties() \
                != other.get_available_data_properties():
            raise ValueError("Available data properties do not match.")
        if self.get_all_prediction_names() \
                != other.get_all_prediction_names():
            raise ValueError("Available prediction names do not match.")

        return TrackPropertiesDataset(
            ndops.add_track_prop_dicts(
                [self._track_prop_dict, other._track_prop_dict]),
            self._label_property,
            other.active_data_properties,
            ndops.add_track_prop_dicts(
                [self._predictions, other._predictions]))

    def __eq__(self, other):
        """Determines whether two TrackPropertiesDatasets have the same
        active properties, label properties, available data, and
        available predictions."""

        return isinstance(self, type(other)) and \
               self.get_label_property() == other.get_label_property() and \
               self.get_active_data_properties() \
               == other.get_active_data_properties() and \
               self.to_track_prop_dict(include_preds=True) \
               == other.to_track_prop_dict(include_preds=True)

    def __ne__(self, other):
        """Returns whether two TrackPropertiesDatasets are unequal,
        defined by negating __eq__."""

        return not self.__eq__(other)

    def get_data_dim(self, just_active_data=True):
        """Returns the dimension of each element in the data portion of
        the dataset. If active is True, return the number of "active"
        degrees of freedom."""

        return len(self.get_active_data_properties()) if just_active_data \
            else len(self.get_available_data_properties())

    def size(self):
        """Returns the number of tracks in this dataset."""

        return len(next(iter(self._track_prop_dict.values())))

    # DATA

    def get_data(self, track_properties=None, normalize=True):
        """Returns data corresponding to the given data properties as
        a tensorflow array. By default, returns non-normalized active
        data.

        Args:
            track_properties: a list of track properties. If None,
                returns the active data.
            normalize: normalize the data within each track property.
                False by default.

        Returns:
            A tensor array of this dataset's data. It is indexed on the
            first axis by track number, and on the second by track
            property.

        Raises:
            ValueError: if one of the given track properties is not in
                this dataset.
        """

        if track_properties is None:
            track_properties = self.get_active_data_properties()
        else:
            for track_property in track_properties:
                if track_property not in self.get_available_data_properties():
                    raise ValueError("Provided track property {} not available"
                                     "in this dataset.".format(track_property))

        return tftrans(tfconst(list(map(lambda track_property:
                    ndops.normalize_val_list(
                        self._track_prop_dict[track_property]) if normalize\
                                else self._track_prop_dict[track_property],
                    track_properties)), dtype=float64))

    def get_active_data_properties(self):
        """Returns a list of the current active data properties in this
        TrackPropertiesDataset."""

        return self._active_data_properties

    def get_available_data_properties(self):
        """Returns a list of all data properties that this dataset has
        available. Note that the label property is an available data
        property."""

        return list(self._track_prop_dict.keys())

    def set_active_data_properties(self, track_properties=None):
        """Sets the given track properties as the active data properties
        within this dataset. Note that the label property is not barred
        from being a data property.

        Args:
            track_properties: a list of track properties, or None. If
                None, all available track properties become active data
                properties.

        Raises:
            ValueError: if this dataset does not have a data property
                corresponding to each element of the given list.
        """

        for track_property in track_properties:
            if track_property not in self.get_available_data_properties():
                raise ValueError("Provided track property {} not available"
                                 "in this dataset.".format(track_property))

        self._active_data_properties = track_properties

    # LABELS

    def get_labels(self):
        """Return a tensor list of this dataset's labels."""

        return tfconst(self._track_prop_dict[self._label_property])

    def get_label_property(self):
        """Return this dataset's label property."""

        return self._label_property

    def set_label_property(self, track_property):
        """Sets this dataset's label property.

        Args:
            track_property: a track property.

        Raises:
            ValueError: if the given track property is not in this
                dataset.
        """

        if track_property not in self.get_available_data_properties():
            raise ValueError("Provided track property " + track_property +
                             "not found in this dataset.")

        self._label_property = track_property

    # PREDICTIONS

    def get_all_predictions(self):
        """Returns a dict from label names to lists of predicted
        labels."""

        return dict(map(lambda pred_name:
                        (pred_name, self.get_prediction(pred_name)),
                        self.get_all_prediction_names()))

    def get_all_prediction_names(self):
        """Returns a list of names of all predictions in this
        dataset."""

        return list(self._predictions.keys())

    def get_prediction(self, pred_name):
        """Returns a list of predicted labels as a list.

        Args:
            pred_name: the name of a prediction.

        Returns:
            A list of predicted labels.

        Raises:
            ValueError: if the given pred_name does not correspond to a
                prediction in this TrackPropertiesDataset.
        """

        if pred_name not in self.get_all_prediction_names():
            raise ValueError("{} is not a prediction in this dataset."
                             .format(pred_name))

        return self._predictions[pred_name]

    def add_prediction(self, pred_name, pred_labels):
        """Adds a list of predictions to this model accessible by a
        name. The predictions must be generated from the data already
        present.

        Args:
            pred_name: a name by which to reference these predictions.
            pred_labels: the predicted labels. They can be probablistic
                or one hot, but some plotting functionality will only
                work for probablistic predictions.
        """

        if len(pred_labels) != self.size():
            raise ValueError("Given prediction list has different number of "
                             "elements ({}) than the number in this dataset ({})"
                             .format(len(pred_labels), len(self.size())))

        self._predictions.update({pred_name: pred_labels})

    def remove_prediction(self, pred_name):
        """Removes a prediction from this dataset's prediction dict."""

        self._predictions.pop(pred_name)

    # CUTS

    def cut(self, selector_dict):
        """Cuts this dataset by track property values, returns the cut
        dataset.

        Args:
            selector_dict: a track property dict selector.

        Returns:
            A cut TrackPropertiesDataset.
        """

        indices_to_cut = ndops.select_indices(
            self._track_prop_dict, selector_dict)
        return TrackPropertiesDataset(
            ndops.cut_track_prop_dict_by_indices(
                self._track_prop_dict, indices_to_cut),
            self.get_label_property(),
            self.get_active_data_properties(),
            ndops.cut_track_prop_dict_by_indices(
                self._predictions, indices_to_cut))

    def split(self, split_list):
        """Returns datasets of number and relative sizes of elements as
        specified in split_dist. Retains the same active properties and
        label property. Does not alter the calling dataset, includes
        predictions.

        Args:
            split_list: a list of relative sizes of output datasets.
                For example, [.7 .2 .1] and [70 20 10] will produce the
                same output.

        Returns:
            A list of TrackPropertiesDatasets.
        """

        split_tpds = ndops.split_track_prop_dict(self._track_prop_dict,
                                                 split_list)

        split_preds = ndops.split_track_prop_dict(self._predictions,
                                                  split_list)

        return list(map(lambda split_tpd, split_preds:
                        TrackPropertiesDataset(split_tpd,
                                               self.get_label_property(),
                                               self.get_active_data_properties(),
                                               split_preds),
                        split_tpds, split_preds))

    # OTHER

    def to_track_prop_dict(self, include_preds=False):
        """Converts this TrackPropertiesDataset to a track properties
        dict, with or without labeled predictions. Includes all
        available track properties, not just active data properties.

        Args:
            include_preds: if True, include predictions in the returned
            track properties dict.

        Returns:
            A track properties dict.
        """

        if include_preds:
            tpd_to_return = deepcopy(self._track_prop_dict)
            tpd_to_return.update(deepcopy(self._predictions))
            return tpd_to_return
        else:
            return deepcopy(self._track_prop_dict)



"""PREDICT: make and measure predictions on discriminator behavior."""

from .. import operations as ndops
from .. import analyze as ndanl
from ..operations import select as sel


def check_pred_labels_size(labels, pred_labels):
    """Raises a ValueError is the predicted labels are a different size
    than the actual labels. Otherwise, does nothing."""

    if len(labels) != len(pred_labels):
        raise ValueError("Predicted labels size differs from labels size")


def apply_threshold(pred_prob_labels, threshold):
    """Sends every prediction in the list below the threshold
    (exclusive) to zero and everything above it (inclusive) to one.
    In the parlance of this file, turns predicted probablistic labels
    into predicted labels. Returns a list."""

    return list(map(lambda pred: 1 if pred >= threshold else 0,
        pred_prob_labels))


def pred_proportion_given_truth_case(labels, pred_labels,
        label_restriction, pred_label_case, threshold=.6):
    """Look at the relative proportion of a value of the predicted
    probability labels, looking only at values who match to an acutal
    label of a particular case.

    This is the generalization of true and false positive rates.

    Args:
        labels: a list or tensor of binary classifer labels.
        pred_labels: a list of predicted binary classifier labels,
            OR a list of probablistic predictions to be converted to
            exact predictions using the threshold.
        label_restriction: a function returning true for values of the
            label case to which the domain should be restricted.
        pred_label_case: a function returning true for values to count
            part of the proportion in the prediction with restricted
            domain.
        threshold: a threshold to apply to the probablistic data
            before computing the agreement. Assumes binary classifier.

    Returns:
        The proportion of predicted values meeting a certain case given
        a restriction of true values meeting a certain case, and the
        error in prediction.

    Raises:
        ValueError: if the true and predicted labels differ in size.
    """

    check_pred_labels_size(labels, pred_labels)

    labels_meet_restriction = list(map(lambda label:
        bool(label_restriction(label)),
        labels if isinstance(labels, list) else labels.numpy()))
    pred_labels_meet_case = list(map(pred_label_case,
        apply_threshold(pred_labels, threshold)))

    domain_size = sum(labels_meet_restriction)
    if domain_size == 0:
        return 0, 0

    num_pred_labels_meet_case_in_domain = sum(map(
        lambda label_meets_restriction, pred_label_meets_case:
        label_meets_restriction and pred_label_meets_case,
        labels_meet_restriction, pred_labels_meet_case))

    return num_pred_labels_meet_case_in_domain / domain_size, ndanl.pred_error(
            domain_size, num_pred_labels_meet_case_in_domain)


def true_positive_rate(labels, pred_labels, threshold=.6):
    """For a binary classifier label, returns the proportion of "true"
    cases that the model predicted correctly. Throws an error if the
    lists are of different sizes.

    Args:
        labels: a list of binary classifier labels.
        pred_labels: a list of predicted binary classifier labels,
            OR a list of probablistic predictions to be converted to
            exact predictions using the threshold.
        threshold: a threshold to apply to the probablistic data.

    Returns:
        The proportion of "true" cases that a model predicted correctly
        and the prediction error.

    Raises:
        ValueError: if the true and predicted labels differ in size.
    """

    return pred_proportion_given_truth_case(labels, pred_labels,
            sel(1), sel(1), threshold)


def false_positive_rate(labels, pred_labels, threshold=.6):
    """For a binary classifier label, returns the proportion of "false"
    cases that the model predicted were "true". Raises an error if the
    lists are of different sizes.

    Args:
        labels: a list of binary classifier labels.
        pred_labels: a list of predicted binary classifier labels,
            OR a list of probablistic predictions to be converted to
            exact predictions using the threshold.
        threshold: a threshold to apply to the probablistic data.

    Returns:
        The proportion of "false" cases that a model predicted "true"
        and the error in prediction.

    Raises:
        ValueError: if the true and predicted labels differ in size.
    """

    return pred_proportion_given_truth_case(labels, pred_labels,
            sel(0), sel(1), threshold)


def predict_labels(model, data):
    """Run the model on each element of a dataset and produce a list of
    probabilistic predictions (note: not logits). Assumes a binary
    classifier. Does not apply a threshold.

    Args:
        model: a tensorflow or sklearn model capable of prediction.
        data: an array of elements that the model can use to make
            predictions.

    Returns:
        A Python list of probabilistic predictions.
    """

    # Different models predict in different ways
    if "keras" in str(type(model)):
        pred_prob_labels = map(lambda l: l[0], model.predict(data))
    else:
        pred_prob_labels = map(lambda l: l[1], model.predict_proba(data))

    return list(pred_prob_labels)


def predict_labels_cuts(tpd_selector, dataset):
    """Return a list of labels corresponding to which tracks were
    selected in a dataset. Interpreted in the context of this ML
    package as predicting some binary track property based on cuts.

    Args:
        tpd_selector: a selector for a track properties dict.
        dataset: a TrackPropertiesDataset.

    Returns:
        A Python list of labels corresponding to which tracks were
        selected in a dataset.
    """

    track_prop_dict = dataset.to_track_prop_dict()

    cut_indices = ndops.select_indices(track_prop_dict, tpd_selector)

    pred_labels = [1 for _ in range(
        ndops.track_prop_dict_length(track_prop_dict))]
    for cut_index in cut_indices:
        pred_labels[cut_index] = 0

    return pred_labels




from tensorflow.keras import Sequential
from tensorflow.keras.layers import Dense
from sklearn.ensemble import GradientBoostingClassifier


def make_neuralnet(train_dataset, eval_dataset=None,
                   hidden_layers=[], epochs=10, classifier_order=1):
    """Makes a neural net in tensorflow using training data and optional
    validation data. Takes in the dimension of the data, has the number
    of specified hidden layers, and ends with probablistic output.

    Args:
        train_dataset: a TrackPropertiesDataset that the model will
            train on.
        eval_dataset: a TrackPropertiesDataset that the model will
            use to evalutate performance at the end of each epoch.
        hidden_layers: a list of hidden layer sizes. By default, there
            are no hidden layers, just the input and the output, which
            are predetermined by data dimension.
        epochs: how many time the neural net is trained on data.
        classifier_order: order of categorization. This is set to 1 at
            default, assuming a binary classifier.

    Returns:
        A trained tensorflow neural net.
    """

    # Build the scaffolding
    linear_model = Sequential()
    hidden_layer_sizes = iter(hidden_layers)
    linear_model.add(Dense(next(hidden_layer_sizes),
                           input_dim=train_dataset.get_data_dim()))
    for layer_size in hidden_layer_sizes:
        linear_model.add(Dense(layer_size, activation="relu"))
    linear_model.add(Dense(classifier_order, activation="sigmoid"))

    # Compile
    linear_model.compile(loss='binary_crossentropy',
                         optimizer='adam',
                         metrics=['accuracy'])

    # Train loop
    steps_per_epoch = train_dataset.size() / epochs
    validation_data = None if eval_dataset is None \
        else (eval_dataset.get_data(), eval_dataset.get_labels())
    linear_model.fit(train_dataset.get_data(), train_dataset.get_labels(),
                     validation_data=validation_data,
                     steps_per_epoch=steps_per_epoch,
                     epochs=epochs,
                     verbose=True)

    return linear_model


def make_gbdt(train_dataset, n_estimators=100, max_depth=3, random_state=23):
    """Make a gradient boosted decision tree in sklearn using training
    data, using Claire's model as reference for creation parameters.

    Args:
        train_dataset: a TrackPropertiesDataset that the model will
            train on.
        n_estimators, max_depth, random_state: check out the sklearn
            documentation for a GradientBoostingClassifier.

    Returns:
        A trained sklearn gradient boosted decision tree.
    """

    gbdt = GradientBoostingClassifier(
        n_estimators=n_estimators,
        max_depth=max_depth,
        random_state=random_state)
    gbdt.fit(train_dataset.get_data().numpy(),
             train_dataset.get_labels().numpy())

    return gbdt



"""Basic operations on ntuple dicts and track property dicts."""

from random import shuffle
from random import seed as set_seed
from copy import deepcopy
from functools import reduce
from math import inf
from warnings import warn
from numpy import cumsum
from numpy import array
from numpy import delete
from numpy import where


def add_ntuple_dicts(ntuple_dicts):
    """Adds together multiple ntuple dicts of with the same track types
    and track type properties. Raises an exception if the dicts do not
    have this "sameness" property.

    Args:
        ntuple_dicts: a list of ntuple dicts with the same track types
            and track type properties.

    Returns:
        An ntuple dictionary with the lists of values of each ntuple
        dict in the input list concatenated.
    """

    track_types = iter(next(iter(ntuple_dicts)).keys())

    return dict(map(lambda track_type:
        (track_type, add_track_prop_dicts(
            list(map(lambda ntuple_dict: ntuple_dict[track_type],
                ntuple_dicts)))),
        track_types))


def add_track_prop_dicts(track_prop_dicts):
    """Adds together multiple track properties dicts of with the same
    properties.

    Args:
        track properties_dicts: a list of track properties dicts with
            the same properties.

    Returns:
        A track properties dictionary with the lists of values of each
        track properties dict in the input list concatenated.

    Raises:
        ValueError: if there is no property shared by all of the track
            property dicts.
    """

    def add_two_track_prop_dicts(tp_so_far, tp_to_add):
        """Adds two track properties dicts together as per rules in
        parent function. Returns the sum."""

        props_in_common = set(tp_so_far.keys()).intersection(
                set(tp_to_add.keys()))

        if props_in_common != set(tp_to_add.keys()):
            warn("Track property dicts have differing value lists. "
                    "Will add only properties in common: {}"
                    .format(props_in_common), UserWarning)

        if not len(props_in_common):
            raise ValueError("Track property dicts to add have no properties "
                    "in common.")

        return dict(map(lambda track_property:
            (track_property,
                tp_so_far[track_property] + tp_to_add[track_property]),
            props_in_common))

    return reduce(add_two_track_prop_dicts, track_prop_dicts)


def mix_track_prop_dicts(track_prop_dicts, seed=None):
    """Mixes together multiple track properties dicts with the same
    properties. 'Mixing', in this context, means to cut all other dicts
    in the list down to the size of the smallest, and then shuffle them
    all together.

    Args:
        track properties_dicts: a list of track properties dicts with
            the same properties.
        seed: a seed for the random shuffling for reproducability.

    Returns:
        A track properties dictionary containing an equal amount of
        tracks from each track properties dictionary in the input list.

    Raises:
        ValueError: if there is no property shared by all of the track
            property dicts.
    """

    min_tpd_size = min(map(track_prop_dict_length, track_prop_dicts))

    return shuffle_track_prop_dict(add_track_prop_dicts(map(
            lambda tpd: reduce_track_prop_dict(tpd, min_tpd_size, seed),
            track_prop_dicts)),
        seed=seed)


def ntuple_dict_length(ntuple_dict):
    """Returns a dictionary from track types to the number of tracks of
    that type. Raises an exception of any value lists within one of its
    track properties dicts are different lengths."""

    return dict(map(lambda track_type, track_prop_dict:
        (track_type, track_prop_dict_length(track_prop_dict)),
        ntuple_dict.keys(), ntuple_dict.values()))


def track_prop_dict_length(track_prop_dict):
    """Returns the number of tracks in a track properties dictionary.
    Raises an exception if the value lists in the input dictionary are
    not all of the same length. Returns zero if the track properties
    dict is empty."""

    # A fancy way of checking if all value lists are the same length
    val_list_lengths = set(map(len, track_prop_dict.values()))
    if len(val_list_lengths) > 1:
        raise ValueError("Invalid track prop dictionary: "
                "value lists are of different sizes")
    if len(val_list_lengths) == 0:
        return 0

    return next(iter(val_list_lengths))


class TrackPropertyDictIterator:
    """Iterates through tracks in a track properties dict, where each
    track is represented as a dictionary from a value name to a single
    property. Does not alter the original track properties dict."""

    def __init__(self, track_prop_dict, increment=1):
        self.tpd = deepcopy(track_prop_dict)
        self.track_index = -1
        self.increment = increment
        self.num_tracks = track_prop_dict_length(track_prop_dict)

    def __iter__(self):
        return self

    def __next__(self):
        self.track_index += self.increment
        if self.track_index < self.num_tracks:
            return self._grab_track_by_index(self.tpd, self.track_index)
        raise StopIteration

    def _grab_track_by_index(self, track_prop_dict, track_index):
        """Returns a dictionary from track properties to single values
        by selecting that value from each list in the track properties
        dict."""

        return dict(map(lambda property_name, val_list:
            (property_name, val_list[track_index]),
            track_prop_dict.keys(), track_prop_dict.values()))


def shuffle_ntuple_dict(ntuple_dict, seed=None):
    """Returns an ntuple dict whose value lists have been shuffled. To
    preserve association between them, value lists of trk and matchtp
    as well as ones for tp and matchtrk have been shuffled in the same
    way.

    Args:
        ntuple_dict: an ntuple dictionary.
        seed: a seed for the random shuffling for reproducability.

    Returns:
        An ntuple dict with its value lists shuffled, preserving the
        association between complementary track types.
    """

    # Generate shuffled indices dictionary
    ntuple_dict_num_tracks = ntuple_dict_length(ntuple_dict)
    shuffled_indices_dict = {"trk": [], "matchtrk": [], "tp": [], "matchtp": []}
    set_seed(seed)

    def generate_shuffled_indices_dict_pair(track_type, track_prop_dict):
        """Generates a pair to be used in the construction of a
        shuffled indices dictionary."""

        tpd_indices = list(range(track_prop_dict_length(track_prop_dict)))
        shuffle(tpd_indices)

        return track_type, tpd_indices

    shuffled_indices_dict.update(dict(map(generate_shuffled_indices_dict_pair,
        ntuple_dict.keys(), ntuple_dict.values())))

    # Ensure that the ntuple dict num tracks dict has the appropriate
    # number of keys
    ntuple_dict_num_tracks.update(dict(map(lambda track_type, indices:
        (track_type, len(indices)),
        shuffled_indices_dict.keys(), shuffled_indices_dict.values())))

    # Ensure that same-length, complementary track types shuffle the same
    if ntuple_dict_num_tracks["trk"] == ntuple_dict_num_tracks["matchtp"]:
        shuffled_indices_dict["trk"] = shuffled_indices_dict["matchtp"]
    if ntuple_dict_num_tracks["matchtrk"] == ntuple_dict_num_tracks["tp"]:
        shuffled_indices_dict["matchtrk"] = shuffled_indices_dict["tp"]

    return dict(map(lambda track_type, track_prop_dict:
        (track_type, shuffle_track_prop_dict(
            track_prop_dict, shuffled_indices_dict[track_type], seed)),
        ntuple_dict.keys(), ntuple_dict.values()))


def shuffle_track_prop_dict(track_prop_dict, shuffled_indices=None, seed=None):
    """Returns a track properties dict whose value lists have been
    shuffled.

    Args:
        track_prop_dict: a track properties dictionary.
        shuffled_indices: a complete list of indices in the range of
            the number of tracks in this track properties dict. Used
            to completely determine a shuffling.
        seed: a seed for the random shuffling for reproducability.

    Returns:
        A track properties dict whose value lists have been shuffled.

    Raises:
        ValueError: if shuffled_indices is different length than
        track_prop_dict.
    """

    def generate_shuffled_indices(tpd_length):
        """Generates a list of shuffled indices for use in shuffling
        tracks in this track property dictionary."""

        tpd_indices = list(range(tpd_length))
        set_seed(seed)
        shuffle(tpd_indices)

        return tpd_indices

    def shuffle_val_list(val_list, shuffled_indices):
        """Shuffles a value list depending on whether there are shuffled
        indices or a random seed provided."""

        return list(map(lambda i: val_list[i], shuffled_indices))

    tpd_length = track_prop_dict_length(track_prop_dict)

    if shuffled_indices is None:
        shuffled_indices = generate_shuffled_indices(tpd_length)
    if len(shuffled_indices) != tpd_length:
        raise ValueError("shuffled_indices arg length ({}) differs from "
                "track_prop_dict length ({})."
                .format(len(shuffled_indices), tpd_length))

    return dict(map(lambda property_name, val_list:
        (property_name, shuffle_val_list(val_list, shuffled_indices)),
        track_prop_dict.keys(), track_prop_dict.values()))


def reduce_ntuple_dict(ntuple_dict, track_limit,
                       shuffle_tracks=False, seed=None):
    """Reduces an ntuple dictionary to a number of tracks. If number of
    tracks in the ntuple is less than the track limit specified, returns
    all tracks. Can be used for convenient print debugging. Does not
    affect the original ntuple dictionary.

    Args:
        ntuple_dict: an ntuple dict.
        track_limit: number of tracks to retain in each value list. Or,
            an integer that will be expanded into a corresponding
            dictionary.
        shuffle_tracks: if True, shuffles the value lists before
            reducing.
        seed: a seed for the shuffling, for reproducability.

    Returns:
        An ntuple dictionary with track_limit tracks.
    """

    # Get track_limit into correct form if it's an int
    if isinstance(track_limit, int):
        track_limit = dict(map(lambda track_type:
            (track_type, track_limit),
            ntuple_dict.keys()))

    if shuffle_tracks:
        ntuple_dict = shuffle_ntuple_dict(ntuple_dict, seed)

    return dict(map(lambda track_type, track_prop_dict:
        (track_type, reduce_track_prop_dict(
            track_prop_dict, track_limit[track_type], shuffle_tracks=False)),
        ntuple_dict.keys(), ntuple_dict.values()))


def reduce_track_prop_dict(track_prop_dict, track_limit, min_index=0,
                           shuffle_tracks=True, seed=None):
    """Reduces a track properties dictionary such that each of its value
    lists are only a certain length. Does not affect the original track
    property dictionary.

    Args:
        track_prop_dict: a track properties dict.
        track_limit: the maximum length for a value list.
        min_index: the first index to include in the result.
        shuffle_tracks: if True, shuffles the value lists before
            reducing.
        seed: a seed for the shuffling, for reproducability.

    Returns:
        A track properties dictionary with reduced-length value lists.
    """

    if shuffle_tracks:
        track_prop_dict = shuffle_track_prop_dict(track_prop_dict, seed=seed)

    return dict(map(lambda track_prop, track_prop_vals:
        (track_prop, track_prop_vals[min_index:min(track_limit + min_index,
            len(track_prop_vals))]),
        track_prop_dict.keys(), track_prop_dict.values()))


def split_track_prop_dict(track_prop_dict, split_list):
    """Splits a track properties dict into smaller ones according to
    the relative sizes of split_list elements. There is no option to
    shuffle these, as the dict can be shuffled before splitting.

    Args:
        track_prop_dict: a track properties dict.
        split_list: a list of positive values that determine the number
            and relative sizes of the post-split track property dicts.

    Returns:
        A list of track property dicts.
    """

    def get_split_sizes(split_list, num_tracks):
        """Returns the sizes of data by normalizing the provided split
        distribution and mutliplying by the number of tracks in such
        a way that the resulting sizes add up to the original tracks."""

        split_list_total = sum(split_list)
        split_sizes = list(map(lambda split_val:
            int(split_val * num_tracks / split_list_total),
            split_list))

        # Ensure the split sizes add up to the total number of tracks
        split_sizes[-1] += num_tracks - sum(split_sizes)

        return split_sizes

    split_boundaries = [0] + list(cumsum(get_split_sizes(split_list,
            track_prop_dict_length(track_prop_dict))))

    return list(map(lambda start_index, end_index:
        reduce_track_prop_dict(track_prop_dict, end_index - start_index,
            start_index, shuffle_tracks=False),
        split_boundaries[:-1], split_boundaries[1:]))


def select(*selector_key, invert=False):
    """Takes in a selector key and returns a selector that returns true
    for selected values and false for non-selected values. This is how
    cuts are applied in ntupledicts.

    Args:
        selector_key: If a single number, the selector will return true
            for that number. If two numbers, the selector will return
            true for numbers in that range, inclusive. If a list, will
            treat each element of a list as a selector and logical OR
            them together. This is done instead of AND, as AND can
            simply be achieved by applying multiple selectors.
        invert: Invert the selection. False by default.

    Returns:
        A selector, a function that returns true for some values and
        false for all others.

    Raises:
        ValueError: for invalid selector keys.
    """

    if len(selector_key) == 1:
        key_contents = next(iter(selector_key))
        if isinstance(key_contents, list):
            selector = lambda val: any(map(
                lambda sub_selector: sub_selector(val),
                key_contents))
        elif isinstance(key_contents, (float, int)):
            selector = lambda val: val == next(iter(selector_key))
        else:
            raise ValueError("Invalid selector key type: {}."
                    .format(type(key_contents)))
    elif len(selector_key) == 2:
        selector = lambda val: selector_key[0] <= val <= selector_key[1]
    else:
        raise ValueError("Invalid selector key length: {}. Read the docs!"
                         .format(selector_key))

    return lambda val: invert != selector(val)


def cut_ntuple_dict(ntuple_dict, nd_selector):
    """Cuts an ntuple dictionary by cutting each track type according to
    a selector dictionary, cutting those tracks not selected. Tracks are
    cut "symmetrically" across corresponding groups, meaning that any
    cuts applied to trks are applied to matchtps, and from tps to
    matchtrks, and vice versa.

    Args:
        ntuple_dict: an ntuple dictionary
        nd_selector: a selector for an ntuple dict

    Returns:
        A cut ntuple dictionary
    """

    # Build list of tracks to cut from tp/matchtrk group and trk/matchtp groups
    cut_indices_dict = {"trk": [], "matchtrk": [], "tp": [], "matchtp": []}
    cut_indices_dict.update(dict(map(lambda track_type, cut_dict:
                                     (track_type, select_indices(
                                         ntuple_dict[track_type], cut_dict)),
                                     nd_selector.keys(),
                                     nd_selector.values())))

    # Combine trk and matchtp, tp and matchtrk indices
    # Sort and remove duplicates
    trk_matchtp_indices_to_cut = sorted(
        list(dict.fromkeys(cut_indices_dict["trk"] +\
                cut_indices_dict["matchtp"])))
    tp_matchtrk_indices_to_cut = sorted(
        list(dict.fromkeys(cut_indices_dict["tp"] +\
                cut_indices_dict["matchtrk"])))

    cut_ntuple_dict = {}
    for track_type in ntuple_dict.keys():
        if track_type in ["trk", "matchtp"]:
            indices_to_cut = trk_matchtp_indices_to_cut
        if track_type in ["tp", "matchtrk"]:
            indices_to_cut = tp_matchtrk_indices_to_cut
        cut_ntuple_dict[track_type] = cut_track_prop_dict_by_indices(
            ntuple_dict[track_type], indices_to_cut)

    return cut_ntuple_dict


def cut_track_prop_dict(track_prop_dict, tpd_selector):
    """Cuts an track properties dictionary by cutting each track type
    according to a cut dictionary.

    Args:
        track_prop_dict: a tracks properties dictionary.
        tpd_selector: a selector for a tracks properties dictionary.

    Returns:
        A cut tracks properties dictionary.
    """

    return cut_track_prop_dict_by_indices(track_prop_dict,
                select_indices(track_prop_dict, tpd_selector))


def select_indices(track_prop_dict, tpd_selector, invert=True):
    """Selects indices from a tracks properties dictionary that meet the
    conditions of the selector dictionary. If a property is in the
    selector dict but not in the tracks properties dict, the program
    won't raise an exception, but will print a message.

    Args:
        track_prop_dict: a tracks properties dictionary.
        tpd_selector: a dictionary from track property names to
            selectors.
        invert: return all indices NOT selected. Default is True. This
            jibes with how this function is mainly used: track cuts.

    Returns:
        Indices from the track properties dict selected by the selector
        dict.
    """

    # Determine which selection conditions will be applied
    for track_property in list(tpd_selector.keys()):
        if track_property not in track_prop_dict.keys():
            warn("{} not in tracks properties; will not select"
                    .format(track_property), UserWarning)
            tpd_selector.pop(track_property)

    tpd_length = track_prop_dict_length(track_prop_dict)
    return list(set(sum(map(lambda track_property, selector:
                            list(where([invert != selector(val) for val in\
                                track_prop_dict[track_property]])[0]),
                            tpd_selector.keys(), tpd_selector.values()),
                        [])))


def cut_track_prop_dict_by_indices(track_prop_dict, indices_to_cut):
    """Takes in a list of indices to cut and cuts those indices from the
    lists of the dictionary. Assumes that all lists in track_prop_dict
    are the same size. This list of indices will frequently be generated
    using get_indices_meeting_condition. The list of indices does not
    have to be sorted by size.

    Args:
        track_prop_dict: a tracks properties dictionary.
        indices_to_cut: a collection of indices to cut. Repeats are
            tolerated, but out-of-range indices will result in an
            exception.

    Returns:
        The same tracks properties dictionary with the given indices
        on its value lists removed.
    """

    track_properties = track_prop_dict.keys()
    post_cuts_track_prop_dict = {}
    for track_property in track_properties:
        post_cuts_track_prop_dict[track_property] = list(delete(
                array(track_prop_dict[track_property]), indices_to_cut))

    return post_cuts_track_prop_dict


def normalize_ntuple_dict(ntuple_dict, normalize_dict=None):
    """Normalizes each value list in an ntuple dict. Does not attempt
    to normalize values of the same property but different track types
    in the same way.

    Args:
        ntuple_dict: an ntuple dict.
        normalize_dict: a dictionary from track types to lists of
            properties to normalize of that track type. If None,
            normalizes all value lists.

    Returns:
        An ntuple dict with the appropriate value lists normalized.
    """

    base_normalize_dict = dict(map(lambda track_type: (track_type, None),
        ntuple_dict.keys()))
    base_normalize_dict.update(normalize_dict)

    return dict(map(lambda track_type:
        (track_type, normalize_track_prop_dict(ntuple_dict[track_type],
            base_normalize_dict[track_type])),
        base_normalize_dict.keys()))


def normalize_track_prop_dict(track_prop_dict, props_to_normalize=None):
    """Returns a track prop dict of the same form as the original, but
    each value list has been divided by its highest value. All values
    are normalized by default, but only some will be normalized if a
    list is given."""

    if props_to_normalize is None:
        props_to_normalize = list(track_prop_dict.keys())
    else:
        for track_property in props_to_normalize:
            if track_property not in list(track_prop_dict.keys()):
                warn("{} not in tracks properties; will not normalize"
                        .format(track_property), UserWarning)

    return dict(map(lambda track_property:
        (track_property, normalize_val_list(track_prop_dict[track_property])),
        props_to_normalize))


def normalize_val_list(val_list):
    """Returns a list of numeric values by the size of their maximum
    value."""

    max_val = float(max(val_list))

    return [ val/max_val if max_val != 0 else 0 for val in val_list ]




from setuptools import setup
from setuptools import find_packages

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="ntupledicts",
    version="0.9.3",
    author="Casey Pancoast",
    author_email="cqpancoast@gmail.com",
    description="Treating CMS TrackTrigger ROOT Ntuples as Python dictionaries with ML studies in mind.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cqpancoast/ntupledicts",
    license="MIT",
    keywords="cms tracktrigger track-trigger root ntuple python dictionary dict ml",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    install_requires=[
        "tensorflow>=2",
        "scikit-learn>=0.22",
        "uproot>=3",
        "matplotlib>=3",
        "numpy>=1"
        ],
    classifiers=[
        "Programming Language :: Python",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Physics"
    ]
)



"""Analyzes the contents of ntuple dicts, track property dicts, and
value lists.

Do things like get the efficiency of an ntuple dict, bin values and
take a measure on each set of binned values, and create custom value
lists that wouldn't be found in the original ntuple.

Also contains a function for finding the error of a prediction given
the prediction and the real values. This can be a prediction as per
efficiency of track finding or a prediction by an ML model.
"""

from . import operations as ndops
from .operations import select as sel
from numpy import linspace
from math import sqrt
from statistics import stdev


def get_proportion_selected(val_list, selector, norm=True):
    """Find the proportion of tracks selected with the given selector.
    If there are no tracks in the tracks property value list, returns
    zero. Can also return the number of tracks meeting the condition.

    Args:
        val_list: a list of values of a track property, such as
            tp_pt or trk_chi2rphi.
        selector: a property that these value can satisfy. For
            example, "lambda trk_eta: trk_eta <= 2.4".
        norm: if True, divides the number of tracks meeting the
            condition by the total number of tracks. This is the default
            option.

    Returns:
        Either the number or proportion of tracks meeting the condition,
        depending on the value of norm.
    """

    if len(val_list) == 0:
        return 0

    num_tracks_meeting_cond = sum(map(selector, val_list))
    return float(num_tracks_meeting_cond) / len(val_list) if norm \
            else num_tracks_meeting_cond


def make_bins(bin_specifier, binning_values):
    """Takes in a bin specifier, which is either an integer number of
    bins, a tuple of the form (lower_bound, upper_bound, num_bins) or
    a list of values, with the last element being the upper bound of the
    last bin.

    If bin_specifier is an integer, it uses the max and min values of
    binned_property to find its range.

    If bin_specifier is a 3-tuple, it creates the third argument number
    of evenly spaced bins between the first two values.

    If bin_specifier is a list, return the list.

    Args:
        bin_specifier: either an int for the number of bins, a 3-tuple
            of the form (low_bound, high_bound, num_bins), or a list of
            numbers
        binning_values: a list of values forming the basis for the bins

    Returns:
        A list of bin edges, of length one greater than the number of
        bins.

    Raises:
        ValueError if bin_specifier is not an int, tuple, or list
    """

    if isinstance(bin_specifier, int):
        bin_specifier = (min(binning_values), max(binning_values),
                bin_specifier)
    if isinstance(bin_specifier, tuple):
        bin_specifier = list(bin_specifier)
        bin_specifier[2] += 1  # we'll need one more value than we want bins
        bin_specifier = list(linspace(*bin_specifier))
    if isinstance(bin_specifier, list):
        return bin_specifier

    raise ValueError("Expected int, tuple, or list as arg 'bin_specifier', "
            "but received {}.".format(str(bin_specifier)))


def take_measure_by_bin(track_prop_dict, bin_property, measure, bins=30):
    """Bin a track properties dict by a value list of a corresponding
    property, then compute some measure for the values in each bin. For
    example, the track_prop_dict could could be of tracking particles
    and contain nmatch, and the measure could be
    eff_from_track_prop_dict.

    Args:
        track_prop_dict: a track properties dict.
        bin_property: a property in track_prop_dict that will split it
            into bins. Preferably a continuous value, but no hard
            restriction is made in this code.
        measure: a function that takes in a track properties dict and
            returns a number and an error.
        bins: either an int for the number of bins, a 3-tuple of the
            form (low_bound, high_bound, num_bins), or a list of
            numbers. See ntupledict.operations.make_bins() for info.

    Returns:
        The bins, bin heights, and errors computed from the binned value
        lists.
    """

    binning_val_list = track_prop_dict[bin_property]
    bins = make_bins(bins, binning_val_list)

    # Sort values into bins with respect to binning value
    bin_heights_and_errs = list(map(lambda lower_bin, upper_bin:
        measure(ndops.cut_track_prop_dict(track_prop_dict,
            # Select values in range lower_bin to upper_bin,
            # but exclude values equal to upper_bin
            {bin_property: lambda val: lower_bin <= val < upper_bin})),
        bins[:-1], bins[1:]))

    bin_heights = list(map(lambda l: l[0], bin_heights_and_errs))
    bin_errs = list(map(lambda l: l[1], bin_heights_and_errs))

    return bins, bin_heights, bin_errs


def pred_error(domain_size, num_selected):
    """Finds the error of a prediction in some domain given the size of
    the domain and the number of correct predictions in that domain. If
    at any point division by zero is attempted, return 0."""

    try:
        return 1 / (domain_size * sqrt(
            num_selected * (1 - (num_selected / domain_size))))
    except ZeroDivisionError:
        return 0


def eff_from_ntuple_dict(ntuple_dict, tp_selector_dict=None):
    """Finds the efficieny of an ntuple dict and its standard deviation.
    Restrictions can be made on the tracking particles by performing a
    cut on the ntuple. Note that the ntuple must contain pt.

    Args:
        ntuple_dict: an ntuple dictionary containing a tracking
            particle track property dict.
        tp_selector_dict: a dictionary from tp properties
            ("pt", "eta", etc.) to conditions (lambda pt: pt < 2, etc.).

    Returns:
        A tuple containing the efficiency of the tracking algorithm for
        the tracks in the given ntuple dict and the standard deviation.
    """

    return eff_from_track_prop_dict(ntuple_dict["tp"], tp_selector_dict)


def eff_from_track_prop_dict(track_prop_dict_tp, selector_dict=None):
    """Finds the efficieny with pred error of an track properties dict.
    Restrictions can be made on the tracking particles by performing a
    cut. Note that the track properties dictionary must be of tracking
    particles.

    Args:
        track_prop_dict_tp: a tracks properties dict carrying value
            lists from tracking particles.
        selector_dict: a dictionary from tp properties
            ("pt", "eta", etc.) to conditions (lambda pt: pt < 2, etc.).

    Returns:
        A tuple containing the efficiency of the tracking algorithm for
        the tracks in the given ntuple dict and the standard deviation.
    """

    if selector_dict is not None:
        track_prop_dict_tps = ndops.cut_track_prop_dict(
                track_prop_dict_tp, selector_dict)

    num_tps = ndops.track_prop_dict_length(track_prop_dict_tps)
    num_matched_tps = num_tps - track_prop_dict_tps["nmatch"].count(0)

    return num_matched_tps / num_tps, pred_error(num_tps, num_matched_tps)


class StubInfo(object):
    """Converts eta and hitpattern into data about stubs for a single
    track.

    The only directly accessible info from this class are boolean
    lists, all of which are indexed by layer/disk:
        - indices 0 - 5 in the lists correspond to layers 1 - 6.
        - indices 6 - 10 in the list correspond to disks 1 - 5.
    Any information you could want about stubs can be found from these
    three lists, sum(), map(), and lambda.

    For example, if I wanted to find the number of missing 2S layers:

        def missing_2S_layers(stub_info):
            return sum(map(lambda expected, hit, ps_2s:
                            not ps_layer and expected and not hit,
                            stub_info.get_expected(),
                            stub_info.get_hit(),
                            stub_info.get_ps_2s()))

    Down below, there are convenience functions process_stub_info and
    basic_process_stub_info for processing instances of this class.

    Note that these definitions are in accordance with the expected and
    missed definitions in the TrackTrigger's Kalman filter used to
    originally create hitpattern. One consequence of this is that there
    will never be hit stub that was not expected.
    """

    def __init__(self, eta, hitpattern):
        """Stores expected, hit, and PS (False for 2S) as tuples of
        boolean values."""

        self._gen_expected(abs(eta))
        self._gen_hit(hitpattern)
        self._gen_ps_2s(abs(eta))

    def _gen_expected(self, abseta):
        """Sets a tuple of boolean values indicating whether the
        Kalman filter expects a hit on a layer/disk for some absolute
        eta. If eta is greater than 2.4, the list will be all False.

        Args:
            abseta: the absolute value of a pseudorapitiy measurement
        """

        # eta regions for and indices of expected layers/disks
        eta_regions = [0., 0.2, 0.41, 0.62, 0.9, 1.26, 1.68, 2.08, 2.4]
        num_layers_disks = 11
        layer_maps = [[1,  2,  3,  4,  5,  6],
                      [1,  2,  3,  4,  5,  6],
                      [1,  2,  3,  4,  5,  6],
                      [1,  2,  3,  4,  5,  6],
                      [1,  2,  3,  4,  5,  6],
                      [1,  2,  3,  7,  8,  9, 10],
                      [1,  2,  8,  9, 10, 11],
                      [1,  7,  8,  9, 10, 11]]

        expected_layers = []
        for eta_low, eta_high, layer_map in zip(
                eta_regions[:-1],
                eta_regions[1:],
                layer_maps):
            if eta_low <= abseta <= eta_high:
                expected_layers = layer_map
                break

        self._expected = tuple(map(lambda index: index + 1 in expected_layers,
            range(num_layers_disks)))

    def _gen_hit(self, hitpattern):
        """Generates a tuple of the same form as the expected hits tuple
        using the hitpattern variable and the expected hits list. Each
        True value in this list represents a hit. The _gen_expected()
        method must be run first.

        Args:
            hitpattern: a number that, when in base two, corresponds to
                a list of zeroes or ones that indicate whether each
                layer in a set of six or seven expected layers were hit.
        """

        def gen_hits_iter(hitpattern, num_expected):
            """Return an iterator through hitpattern by converting it
            into a list of boolean values, ordered by ascending
            magnitude in the original hitpattern. Falses are included
            at the end of the list until it is the same length as the
            expected number of values (6 or 7)."""

            hits_bool = [bool(int(i)) for i in bin(hitpattern)[-1:1:-1]]
            return iter(hits_bool + (num_expected - len(hits_bool)) * [False])

        hits_iter = gen_hits_iter(hitpattern, len(self._expected))
        self._hit = tuple(map(lambda expected:
            expected and next(hits_iter),
            self.get_expected()))

    def _gen_ps_2s(self, abseta):
        """Generates a tuple indexed by layer for which each boolean
        value represents whether a layer or disk is PS (True) or 2S
        (False). This is necessary because a given disk has PS and 2S
        modules, separated by eta.

        Args:
            abseta: the absolute value of a pseudorapitiy measurement
        """

        layer_ps_2s = 3 * (True,) + 3 * (False,)

        disk_ps_2s_cuts = [1.45, 1.6, 1.8, 1.975, 2.15]
            # ps above, 2s below
        disk_ps_2s = tuple(map(lambda disk_ps_2s_cut:
            abseta > disk_ps_2s_cut,
            disk_ps_2s_cuts))

        self._ps_2s = layer_ps_2s + disk_ps_2s

    def get_expected(self):
        """Returns a list of booleans representing which layers/disks
        were expected to be hit by the Kalman filter."""

        return list(self._expected)

    def get_hit(self):
        """Returns a list of booleans representing which layers/disks
        were hit, within the layers/disks expected byt the Kalman
        filter."""

        return list(self._hit)

    def get_ps_2s(self):
        """Returns a list of booleans indexed by layer/disk indicating
        if the layer or disk with that index is PS (True) or 2S
        (False)."""

        return list(self._ps_2s)


def create_stub_info_list(track_prop_dict, process_stub_info):
    """Uses eta and hitpattern to generate a list of StubInfos from the
    given track property dict. Then maps those StubInfos to something
    else using some function.

    Args:
        track_prop_dict: a tracks properties dict with track properties
            eta and hitpattern. Must represent either trk or matchtrk,
            as only those have the hitpattern track property.
        process_stub_info: a function or lambda expression that accepts
            StubInfos.

    Returns:
        A list of processed StubInfos indexed by track.
    """

    return list(map(lambda eta, hitpattern:
        process_stub_info(StubInfo(eta, hitpattern)),
        track_prop_dict["eta"], track_prop_dict["hitpattern"]))


def basic_process_stub_info(process_layer):
    """Returns a StubInfo processing function that is agnostic towards
    layer indices, which means it should work for most cases.

    For example, a function that determines how many missing 2S layers
    are in a StubInfo would be:

        basic_process_stub_info(lambda expected, hit, ps_2s:
                                not ps_2s and expected and not hit)

    Args:
        process_layer: A function from a single layer's expected bool,
        hit bool, and ps/2s bool (in that order) to a boolean.

    Returns:
        A function that accepts a StubInfo and counts for how many
        layers process_layer returns True.
    """

    return lambda stub_info: sum(map(process_layer,
                                     stub_info.get_expected(),
                                     stub_info.get_hit(),
                                     stub_info.get_ps_2s()))




import matplotlib.pyplot as plt
from . import analyze as ndanl
from . import operations as ndops
from .operations import select as sel
from math import sqrt


def plot_roc_curve_from_cut_list(ntuple_dict, cut_property, cuts,
                                 cuts_constricting=True, group_name=None,
                                 ax=None):
    """Adjusts the cut on a certain variable in an event set and plots
    the change in tracking efficiency and fake rate.

    Args:
        ntuple_dict: an ntuple dict that contains trk with at least
            genuine and the cut property, matchtrk with at least the
            cut property, and tp with at least nmatch.
        cut_property: the variable to change cuts on.
        cuts: a list of length 2 lists containing lower and upper cut
            bounds (inclusive). These are used rather than selectors so
            that information about the cut can be used in the final
            graph.
        cuts_constricting: if the cuts are strictly in increasing order
            of strictness, the same list can be preserved, decreasing
            runtime. True by default.
        group_name: a legend entry string for this curve to identify
            itself others in an overlay
        ax: an axes object to be used to plot this graph.

    Returns:
        The axes object used to plot this graph.
    """

    if ax is None:
        ax = plt.figure().add_subplot(111)

    # Build up cuts plot info, which is what will be plotted
    cuts_plot_info = {"cut": [], "eff": [], "fake_rate": []}
    cut_ntuple_dict = ntuple_dict
    for cut in cuts:
        # If cuts are constricting, use cut ntuple dict from previous iteration
        ntuple_dict_to_cut = cut_ntuple_dict if cuts_constricting \
            else ntuple_dict
        tpd_selector = {cut_property: sel(*cut)}
        cut_ntuple_dict = ndops.cut_ntuple_dict(ntuple_dict_to_cut,
                                                {"trk": tpd_selector,
                                                 "matchtrk": tpd_selector})
        cuts_plot_info["cut"].append(cut[1])  # only get upper bound of cut
        cuts_plot_info["eff"].append(ndanl.eff_from_ntuple_dict(
            cut_ntuple_dict["tp"]["pt"]))
        cuts_plot_info["fake_rate"].append(ndanl.get_proportion_selected(
            cut_ntuple_dict["trk"]["genuine"], sel(0)))

    # Now plot!
    ax.plot(cuts_plot_info["fake_rate"], cuts_plot_info["eff"], "b.",
            label=group_name)

    ax.set_xlabel("Track fake rate")
    ax.set_ylabel("Tracking efficiency")
    ax.legend()
    for fake_rate, eff, cut in zip(
            cuts_plot_info["fake_rate"],
            cuts_plot_info["eff"],
            cuts_plot_info["cut"]):
        ax.annotate(str(cut), ax=(fake_rate, eff))

    return ax


def plot_measure_by_bin(track_prop_dict, bin_property, measure,
                        bins=30, legend_id=None, ax=None):
    """Splits a track property dict into bins by some value, then
    computes a measure on each binned track property dict, including
    some error.

    Args:
        track_prop_dict: a track properties dict.
        bin_property: a property in track_prop_dict that will split it
            into bins. Preferably a continuous value, but no hard
            restriction is made in this code.
        measure: a function that takes in a track properties dict and
            returns a numerical value and an error.
        bins: either an int for the number of bins, a 3-tuple of the
            form (low_bound, high_bound, num_bins), or a list of
            numbers. See ntupledict.operations.make_bins() for info.
        legend_id: the entry in the legend for the line to be plotted.
            Calling ax.legend() should be done outside this function.
        ax: an axes object to overlay this data onto a previous plot.

    Returns:
        A matplotlib.pyplot.Axes object for adjusting plot properties
        and overlaying data.
    """

    if ax is None:
        ax = plt.figure().add_subplot(111)

    bins, bin_heights, bin_errs = ndanl.take_measure_by_bin(track_prop_dict,
                                                  bin_property, measure, bins)
    bin_middles = list(map(lambda lower, upper: (lower + upper) / 2,
                           bins[:-1], bins[1:]))

    ax.errorbar(bin_middles, bin_heights, yerr=bin_errs,
            label=legend_id, fmt=".")
    ax.set_xlabel(bin_property)

    return ax


def plot_property_bin_hist(track_prop_dict, track_property, bins=30,
                       legend_id=None, ax=None):
    """Plot a histogram distribution of a track property in a track
    properties dict. Error bars are simply one over the square root of
    the number of occurrences.

    Args:
        track_prop_dict: a track properties dict
        track_property: a property in track_prop_dict that will split it
            into bins. Preferably a continuous value, but no hard
            restriction is made in this code
        bins: either an int for the number of bins, a 3-tuple of the
            form (low_bound, high_bound, num_bins), or a list of
            numbers. See ntupledict.operations.make_bins() for info
        legend_id: the entry in the legend for the line to be plotted.
            Calling ax.legend() should be done outside this function.
        ax: an axes object to overlay this data onto a previous plot

    Returns:
        A matplotlib.pyplot.Axes object for adjusting plot properties
        and overlaying data
    """

    if ax is None:
        ax = plt.figure().add_subplot(111)

    ax = plot_measure_by_bin(track_prop_dict, track_property,
                             lambda tpd: (len(tpd[track_property]),
                                 sqrt(len(tpd[track_property]))),
                             bins, legend_id, ax)
    ax.set_ylabel("num. tracks")
    ax.set_title("Histrogram of {}".format(track_property))

    return ax


def plot_property_scatter(track_prop_dict, track_property, legend_id=None,
        ax=None):
    """Plot a scatter plot of a track property in a track properties
    dict. Intended for track properties who take discrete values. Error
    bars are simply one over the square root of a number of occurrences.

    Args:
        track_prop_dict: a track properties dict
        track_property: a property in track_prop_dict that will split it
            into bins. Preferably a continuous value, but no hard
            restriction is made in this code
        legend_id: the entry in the legend for the line to be plotted.
            Calling ax.legend() should be done outside this function.
        ax: an axes object to overlay this data onto a previous plot

    Returns:
        A matplotlib.pyplot.Axes object for adjusting plot properties
        and overlaying data
    """

    if ax is None:
        ax = plt.figure().add_subplot(111)

    # Generate dict from values to number of occurences
    val_list = track_prop_dict[track_property]
    hist_dict = dict(map(lambda value:
        (str(value), val_list.count(value)),
        set(val_list)))
    hist_yerr = list(map(lambda height: 1 / sqrt(height),
        hist_dict.values()))

    ax.errorbar(hist_dict.keys(), hist_dict.values(), yerr=hist_yerr,
            label=legend_id, fmt=".")
    ax.set_yscale("log")  # we usually want log when we're counting
    ax.set_xlabel(track_property)
    ax.set_ylabel("number of tracks")
    ax.set_title("Histogram of {}".format(track_property))

    return ax

